from django.conf import settings as django_settings
from django.core.exceptions import FieldDoesNotExist
from django.utils.text import capfirst
from django.db.models.fields import NOT_PROVIDED

from rest_framework import serializers, relations
from rest_framework.fields import empty

from .app_settings import settings


def reverse(*args, **kwargs):
    try:
        exception = ModuleNotFoundError
    except:
        ## py 3.6+
        exception = ImportError

    try:
        from django.core.urlresolvers import reverse
    except exception:
        # Django 1.11+
        from django.urls import reverse

    return reverse(*args, **kwargs)


def get_validation_attrs(instance_field):
    rv = {}

    attrs_to_validation = {
        'min_length': 'min',
        'max_length': 'max',
        'min_value': 'min',
        'max_value': 'max'
    }
    for attr_name, validation_name in attrs_to_validation.items():
        if getattr(instance_field, attr_name, None) is not None:
            rv[validation_name] = getattr(instance_field, attr_name)

    return rv


def get_field_dict(field, serializer, translated_fields=None, fields_annotation=False, model=None):
    if translated_fields is None:
        translated_fields = []

    serializer_instance = serializer()
    name = field['name'] if isinstance(field, dict) else field
    try:
        field_instance = serializer_instance.fields[name]
    except KeyError:
        return {'key': name}
    read_only = name == '__str__'
    if not read_only and field_instance.read_only:
        if not isinstance(field_instance, serializers.ManyRelatedField):
            read_only = True

    rv = {
        'key': name,
        'type': settings.WIDGET_MAPPING[field_instance.__class__.__name__],
        'read_only': read_only,
        'ui': {
            'label': name.title().replace('_', ' ')
            if name != '__str__'
            else serializer_instance.Meta.model.__name__,
        },
        'validation': {
            'required': field_instance.required,
        },
        'extra': {},
        'translated': name in translated_fields
    }

    if fields_annotation and name in fields_annotation:
        if 'placeholder' in fields_annotation[name]:
            rv['ui']['placeholder'] = fields_annotation[name]['placeholder']
        if 'help' in fields_annotation[name]:
            rv['ui']['help'] = fields_annotation[name]['help']
    if field_instance.help_text is not None and 'help' not in rv['ui']:
        rv['ui']['help'] = field_instance.help_text

    default = field_instance.default
    model_field = None
    if model:
        try:
            model_field = model._meta.get_field(field_instance.source)
        except FieldDoesNotExist:
            pass

    if default and default != empty:
        rv['default'] = default
    elif default == empty and hasattr(model_field, 'default'):
        default = model_field.default
        if default != NOT_PROVIDED:
            if callable(default):
                default = default()
            rv['default'] = default

    if isinstance(field_instance, (relations.PrimaryKeyRelatedField, relations.ManyRelatedField)):
        if model_field:
            related_model = model_field.related_model
            rv['type'] = settings.WIDGET_MAPPING[model_field.__class__.__name__]
        else:
            related_model = field_instance.queryset.model
        if model_field and model_field.__class__.__name__ == 'ManyToManyRel':
            rv['validation']['required'] = False
        rv['related_endpoint'] = '{}/{}'.format(
            related_model._meta.app_label,
            related_model._meta.model_name.lower()
        )
    elif hasattr(field_instance, 'choices'):
        rv['type'] = settings.WIDGET_MAPPING['choice']
        rv['choices'] = [
            {
                'label': v,
                'value': k,
            } for k, v in field_instance.choices.items()
        ]

    rv['validation'].update(get_validation_attrs(field_instance))

    if isinstance(field, dict):
        extra = rv['extra']
        extra.update(field.get('extra', {}))
        rv.update(field)
        rv['extra'] = extra

    return rv


def action_kwargs(icon_class, btn_class, text, func, kwargs):

    kwargs.update({
        'icon_class': icon_class if icon_class is not None else settings.ACTION_ICON_CLASS,
        'btn_class': btn_class if btn_class is not None else settings.ACTION_BTN_CLASS,
    })

    if text is None:
        kwargs['text'] = capfirst(func.__name__.lower())
    else:
        kwargs['text'] = text

    return kwargs


def get_languages():
    if django_settings.USE_I18N:
        return [
            x[0] for x in getattr(django_settings, 'LANGUAGES', [[django_settings.LANGUAGE_CODE]])
        ]
    else:
        return None



import os
from setuptools import setup

try:
    import pypandoc

    README = pypandoc.convert_file('README.md', 'rst')
except ImportError:
    with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
        README = readme.read()


# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='drf-schema-adapter',
    version='0.9.27',
    packages=['drf_auto_endpoint', 'export_app', 'export_app.management',
              'export_app.management.commands'],
    include_package_data=True,
    license='MIT License',
    description='Making using Django with frontend libraries and frameworks DRYer',
    long_description=README,
    url='https://github.com/drf-forms/drf-schema-adapter',
    author='Emmanuelle Delescolle, Adrien Brunet, Mauro Bianchi, Mattia Larentis, Aaron Elliot Ross',
    author_email='info@levit.be',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=[
        'Django>=1.8',
        'djangorestframework<4.0.0',
        'django-filter>=0.13.0',
        'Inflector>=2.0.11',
    ]
)



import inspect

from django.utils.text import capfirst
from django.utils.module_loading import import_string

from inflector import Inflector

from rest_framework.serializers import PrimaryKeyRelatedField

from .app_settings import settings
from .utils import action_kwargs, get_field_dict, get_languages


def custom_action(method='GET', type='request', icon_class=None, btn_class=None, text=None, **kwargs):

    kwargs.update({
        'type': type,
    })

    def decorator(func):
        func.bind_to_methods = [method, ]
        func.detail = True
        func.action_type = 'custom'
        func.action_kwargs = action_kwargs(icon_class, btn_class, text, func, kwargs)
        func.kwargs = {}

        return func

    return decorator


def bulk_action(method='GET', type='request', icon_class=None, btn_class=None, text=None, **kwargs):

    kwargs.update({
        'type': type,
    })

    def decorator(func):
        func.bind_to_methods = [method, ]
        func.detail = False
        func.action_type = 'bulk'
        func.action_kwargs = action_kwargs(icon_class, btn_class, text, func, kwargs)
        func.action_kwargs['atOnce'] = func.action_kwargs.get('atOnce', True)
        func.kwargs = {}

        return func

    return decorator


def wizard(target_model, serializer, icon_class=None, btn_class=None, text=None, **kwargs):

    inflector_language = import_string(settings.INFLECTOR_LANGUAGE)
    inflector = Inflector(inflector_language)

    _kwargs = {
        'type': 'wizard',
        'params': {},
    }
    _kwargs.update(kwargs)
    kwargs = _kwargs

    fieldsets = kwargs.pop('fieldsets', None)
    if fieldsets is None:
        fieldsets = [
            {'name': field}
            for field in serializer.Meta.fields
        ]
    kwargs['params']['fieldsets'] = fieldsets

    serializer_instance = serializer()
    needs = []
    fields = []
    Adapter = import_string(settings.METADATA_ADAPTER)
    for field in serializer.Meta.fields:
        field_instance = serializer_instance.fields[field]
        if isinstance(field_instance, PrimaryKeyRelatedField):
            model = field_instance.queryset.model
            needs.append({
                'app': model._meta.app_label,
                'singular': model._meta.model_name.lower(),
                'plural': inflector.pluralize(model._meta.model_name.lower()),
            })
        fields.append(Adapter.adapt_field(get_field_dict(field, serializer)))
    kwargs['params']['needs'] = needs
    kwargs['params']['fields'] = fields
    kwargs['languages'] = get_languages()

    def decorator(func):
        # cls = getattr(inspect.getmodule(func),
        #               func.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0])
        func.bind_to_methods = [kwargs.pop('method', 'POST'), ]
        func.detail = True
        func.wizard = True
        func.action_type = 'custom'
        func.action_kwargs = action_kwargs(icon_class, btn_class, text, func, kwargs)
        func.kwargs = {}
        func.action_kwargs['params']['model'] = '{}/{}/{}'.format(
            target_model._meta.app_label.lower(),
            inflector.pluralize(target_model._meta.model_name.lower()),
            func.__name__
        )
        func.serializer = serializer

        return Adapter.adapt_wizard(func)

    return decorator
